__title__ = "multi_realsense_manager"
__version__ = "0.1.7"
__author__ = "DIYer22"
__author_email__ = "ylxx@live.com"
__url__ = "https://github.com/DIYer22/multi_realsense_manager"
__description__ = "Robust high-performance multi-realsenses manager"
